from cspy.algorithms.bidirectional import BiDirectional
from cspy.algorithms.tabu import Tabu
from cspy.algorithms.greedy_elimination import GreedyElim
from cspy.algorithms.psolgent import PSOLGENT
from cspy.algorithms.grasp import GRASP
from cspy.checking import check
from cspy.preprocessing import preprocess_graph

name = "cspy"

__all__ = [
    'BiDirectional', 'Tabu', 'GreedyElim', 'PSOLGENT', 'GRASP', 'check',
    'preprocess_graph'
]

__version__ = '0.1.0'



from operator import add
from collections import deque
from logging import getLogger
from numpy import zeros, array
from types import BuiltinFunctionType
from itertools import filterfalse, tee, chain
from networkx import shortest_simple_paths, NetworkXException

from cspy.checking import check
from cspy.preprocessing import preprocess_graph

log = getLogger(__name__)


class PathBase(object):
    """
    Parent class for :class:Tabu, :class:GRASP, :class:GreedyElim,
    and :class:PSOLGENT.

    Contains path specific functionalitites.
    e.g. feasibility checks, compatible joining
    """

    def __init__(self, G, max_res, min_res, REF=None, preprocess=False):
        # Check inputs
        check(G, max_res, min_res, REF, algorithm=__name__)
        # Preprocess graph
        self.G = preprocess_graph(G, max_res, min_res, preprocess, REF)

        self.max_res = max_res
        self.min_res = min_res
        # Update resource extension function if given
        if REF:
            self.REF = REF
        else:
            self.REF = add

        # Attribute to hold source-sink path
        self.st_path = None

        # Attribrutes for exposure #
        # resource feasible source-sink path
        self.best_path = None
        # Final cost
        self.best_path_cost = None
        # Final resource consumption
        self.best_path_total_res = None

    @property
    def path(self):
        """
        Get list with nodes in calculated path.
        """
        if not self.best_path:
            raise Exception("Please call the .run() method first")
        return self.best_path

    @property
    def total_cost(self):
        """
        Get accumulated cost along the path.
        """
        if not self.best_path:
            raise Exception("Please call the .run() method first")
        return self.best_path_cost

    @property
    def consumed_resources(self):
        """
        Get accumulated resources consumed along the path.
        """
        if not self.best_path:
            raise Exception("Please call the .run() method first")
        return self.best_path_total_res

    def update_simple_path(self, source, max_depth):
        depth = 0

        # Create two copies of the simple path generator
        paths, paths_backup = tee(shortest_simple_paths(self.G, source, 'Sink'),
                                  2)

        try:
            cost_min = 0
            paths_backup = filterfalse(
                lambda p: sum(self.G[i][j]["weight"]
                              for i, j in zip(p, p[1:])) >= 0, paths_backup)
            first = paths_backup.__next__()
            # add first element back
            paths = chain([first], paths_backup)
        except StopIteration:
            # if there exist paths with negative cost
            cost_min = 1e10
        except KeyError:
            return

        for p in paths:
            c = sum(self.G[i][j]["weight"] for i, j in zip(p, p[1:]))
            if c < cost_min:
                path = p
                cost_min = c
            if depth > max_depth:
                return path
            depth += 1
        # No break
        else:
            return path

    def check_feasibility(self, return_edge=True):
        """
        Checks for feasibility for a valid source-sink path.
        If the path (in the st_path attribute) is feasible, then it
        returns true and saves the path, cost and total resources consumed in
        the process.
        Otherwise, by default, it returns the edge that makes the path
        infeasible.
        This can changed with the ``return_edge`` parameter.
        """
        shortest_path_edges = deque(
            e for e in zip(self.st_path, self.st_path[1:]))
        shortest_path_edges_w_data = deque()
        shortest_path_edges_w_data.extend([
            (e[0], e[1], self.G[e[0]][e[1]]) for e in shortest_path_edges
        ])
        # init total resources and cost
        total_res = zeros(self.G.graph['n_res'])
        _cost = 0
        # Check path for resource feasibility by adding one edge at a time
        for edge in shortest_path_edges_w_data:
            _cost += edge[2]['weight']
            if isinstance(self.REF, BuiltinFunctionType):
                total_res += self._edge_extract(edge)
            else:
                total_res = self.REF(total_res, edge)
            if (all(total_res <= self.max_res) and
                    all(total_res >= self.min_res)):
                pass
            else:
                break
        else:
            # Fesible path found. Save attributes.
            self.best_path = self.st_path
            self.best_path_total_res = total_res
            self.best_path_cost = _cost
            return True
        # Return infeasible edge unless specified
        if return_edge:
            return edge
        else:
            return False

    def remove_edge(self, edge):
        if edge[:2] in self.G.edges():
            log.debug("Removed edge {}".format(edge[:2]))
            self.G.remove_edge(*edge[:2])

    def add_edge_back(self, edge):
        log.debug("Added edge back {}".format(edge[:2]))
        if "data" in edge[2]:
            self.G.add_edge(*edge[:2],
                            res_cost=edge[2]['res_cost'],
                            weight=edge[2]['weight'],
                            data=edge[2]['data'])
        else:
            self.G.add_edge(*edge[:2],
                            res_cost=edge[2]['res_cost'],
                            weight=edge[2]['weight'])

    @staticmethod
    def _edge_extract(edge):
        return array(edge[2]['res_cost'])



# TODO: write checks for all inputs

from __future__ import absolute_import
from __future__ import print_function

from math import factorial
from logging import getLogger
from collections import deque
from numpy.random import choice
from random import sample, randint
from itertools import permutations, repeat

# Local imports
from cspy.algorithms.path_base import PathBase

log = getLogger(__name__)


class GRASP(PathBase):
    """
    Greedy Randomised Adaptive Search Procedure for the (resource) constrained
    shortest path problem. Adapted from `Ferone et al 2019`_.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute. Also, the number of nodes must be
        :math:`\geq 5`.

    max_res : list of floats
        :math:`[M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage.

    min_res : list of floats
        :math:`[L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage.

    REF : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default : additive.

    preprocess : bool, optional
        enables preprocessing routine. Default : False.

    max_iter : int, optional
        Maximum number of iterations for algorithm. Default : 100.

    max_localiter : int, optional
        Maximum number of local search iterations. Default : 10.

    alpha : float, optional
        Greediness factor 0 (random) --> 1 (greedy). Default : 0.2.

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs

    Returns
    -------
    path : list
        nodes in resource feasible shortest path obtained.

    Raises
    ------
    Exception
        if no resource feasible path is found

    Notes
    -----
    The input graph must have a ``n_res`` attribute.
    The edges in the graph must all have a ``res_cost`` attribute.
    Also, we must have ``len(min_res)`` :math:`=` ``len(max_res)``.
    See `Using cspy`_

    .. _Using cspy: https://cspy.readthedocs.io/en/latest/how_to.html


    Example
    -------
    To run the algorithm, create a :class:`GRASP` instance and call `run`.

    .. code-block:: python

        >>> from cspy import GRASP
        >>> from networkx import DiGraph
        >>> from numpy import array
        >>> G = DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('Source', 'B', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('Source', 'C', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('A', 'C', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('A', 'E', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('A', 'F', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('B', 'C', res_cost=array([2, 1]), weight=-1)
        >>> G.add_edge('B', 'F', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('B', 'E', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('C', 'D', res_cost=array([1, 1]), weight=-1)
        >>> G.add_edge('D', 'E', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('D', 'F', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('D', 'Sink', res_cost=array([10, 10]), weight=10)
        >>> G.add_edge('F', 'Sink', res_cost=array([10, 1]), weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=array([1, 1]), weight=1)
        >>> grasp = GRASP(G, [5, 5], [0, 0], max_iter=50,
                         max_localiter=10)
        >>> grasp.run()
        >>> path = grasp.path
        >>> print(path)
        ['Source', 'A', 'C', 'D', 'E', 'Sink']

    .. _Ferone et al 2019: https://www.tandfonline.com/doi/full/10.1080/10556788.2018.1548015

    """

    def __init__(self,
                 G,
                 max_res,
                 min_res,
                 REF=None,
                 preprocess=False,
                 max_iter=100,
                 max_localiter=10,
                 alpha=0.2):
        # Pass arguments to parent class
        super().__init__(G, max_res, min_res, REF, preprocess)
        # Algorithm specific attributes
        self.max_iter = max_iter
        self.max_localiter = max_localiter
        self.alpha = alpha
        # Algorithm specific parameters
        self.it = 0
        self.stop = False
        self.best_path = None
        self.best_solution = None
        self.nodes = self.G.nodes()

    def run(self):
        """
        Calculate shortest path with resource constraints.
        """
        while self.it < self.max_iter and not self.stop:
            self._algorithm()
            self.it += 1
        if self.best_solution.path:
            pass
        else:
            raise Exception("No resource feasible path has been found")

    def _algorithm(self):
        solution = self._construct()
        solution = self._local_search(solution)
        self._update_best(solution)

    def _construct(self):
        solution = Solution(sample(self.nodes, 1), 0)  # Init solution
        # Construction phase
        while len(solution.path) < len(self.nodes):
            candidates = [i for i in self.nodes if i not in solution.path]
            weights = deque(
                map(self._heuristic, repeat(solution.path[-1]), candidates))
            # Build Restricted Candidiate List (RCL)
            restriced_candidates = [
                candidates[i]
                for i, c in enumerate(weights)
                if c <= (min(weights) + self.alpha *
                         (max(weights) - min(weights)))
            ]
            # Select random node from RCL to add to the current solution
            solution.path.append(choice(restriced_candidates))
            solution.cost = self._cost_solution(solution)
        return solution

    def _local_search(self, solution):
        it = 0  # init local iteration counter
        while it < self.max_localiter:  # Local search phase
            # Init candidate solution using random valid path generator
            candidate = Solution(
                self._find_alternative_paths(self.G, solution.path), 0)
            # evaluate candidate solution
            candidate.cost = self._cost_solution(candidate)
            # Update solution with candidate if lower cost and resource feasible
            if (candidate.path and candidate.cost < solution.cost and
                    self._check_path(candidate)):
                solution = candidate
            it += 1
        return solution

    def _update_best(self, solution):
        if not self.best_solution or solution.cost < self.best_solution.cost:
            self.best_solution = solution

    def _heuristic(self, i, j):
        # Given a node pair returns a weight to apply
        if i and j:
            if (i, j) not in self.G.edges():
                return 1e10
            else:
                return self.G.get_edge_data(i, j)['weight']
        else:
            return 1e10

    def _cost_solution(self, solution=None):
        if solution:
            return sum(
                self._heuristic(i, j)
                for i, j in zip(solution.path, solution.path[1:]))
        else:
            return 1e11

    def _check_path(self, solution=None):
        """
        Returns True if solution.path is valid and resource feasible,
        False otherwise
        """
        if solution:
            path, cost = solution.path, solution.cost
            if (len(path) > 2 and cost < 1e10 and path[0] == 'Source' and
                    path[-1] == 'Sink'):
                self.st_path = path
                return self.check_feasibility(return_edge=False)
            else:
                return False
        else:
            return False

    @staticmethod
    def _find_alternative_paths(G, path):
        """
        Static Method used in local search to randomly generate valid paths.
        Using a subset of edges, it generates a connected path starting at
        the source node.
        """
        n_permutations = int(factorial(len(path)) / factorial(len(path) - 2))
        sample_size = randint(3, n_permutations)
        selection = sample(deque(permutations(path, 2)), sample_size)
        path_edges = dict(list(edge for edge in selection if edge in G.edges()))
        elem = 'Source'  # start point in the new list
        new_list = []
        for _ in range(len(path_edges)):
            try:
                new_list.append((elem, path_edges[elem]))
                elem = path_edges[elem]
            except KeyError:
                pass
        if new_list:
            nodes_to_keep = [t[0] for t in new_list]
            nodes_to_keep.append(new_list[-1][1])
        else:
            nodes_to_keep = []
        return nodes_to_keep


class Solution(object):
    """
    Object for solutions and candidates during GRASP iterations.

    Parameters
    ----------

    path : list
        list of nodes in current path

    cost : float
        cost of solution
    """

    def __init__(self, path, cost):
        self.path = path
        self.cost = cost



from copy import deepcopy
from logging import getLogger
from networkx import single_source_bellman_ford
from cspy.checking import check

log = getLogger(__name__)


def prune_graph(G, max_res, min_res):
    """
    Graph pruning to remove unreachable nodes.

    Note
    -----
    path_s and path_t contains all partial paths e.g.
        {node_i: [Source, ..., node_k, node_i]}.
    or  {node_i: [Sink, ..., node_k, node_i]}
    Hence, if node_i is found to violate a resource bound, it is
    because of node_k, therefore, we add path[key][-2] = node_k to
    the dictionary of nodes to remove.
    """

    def _check_resource(r):
        # check resource r's feasibility along a path

        def __get_weight(i, j, attr_dict):
            # returns number to use as weight for the algorithm
            return attr_dict['res_cost'][r]

        # Get paths from source to all other nodes
        length_s, path_s = single_source_bellman_ford(G,
                                                      'Source',
                                                      weight=__get_weight)
        length_t, path_t = single_source_bellman_ford(G.reverse(copy=True),
                                                      'Sink',
                                                      weight=__get_weight)
        try:
            # Collect nodes in paths that violate the resource bounds
            # see note above
            nodes_source.update({
                path_s[key][-2]: (val, r)
                for key, val in length_s.items()
                if val > max_res[r] or val < min_res[r]
            })
            nodes_sink.update({
                path_t[key][-2]: (val, r)
                for key, val in length_t.items()
                if val > max_res[r] or val < min_res[r]
            })
        except IndexError:  # No nodes violate resource limits
            pass

    nodes_source = {}
    nodes_sink = {}
    n_nodes = len(G.nodes())
    # Map function for each resource
    list(map(_check_resource, range(0, G.graph['n_res'])))
    if nodes_source and nodes_sink:  # if there are nodes to remove
        # Filter out source or sink
        nodes_to_remove = [
            node for node in G.nodes() if node in nodes_source and nodes_sink
        ]
        if "Source" in nodes_to_remove or "Sink" in nodes_to_remove:
            if "Sink" in nodes_source:
                unreachable_res = nodes_source["Sink"][1]
            elif "Sink" in nodes_sink:
                unreachable_res = nodes_sink["Sink"][1]
            elif "Source" in nodes_sink:
                unreachable_res = nodes_sink["Source"][1]
            elif "Source" in nodes_source:
                unreachable_res = nodes_source["Source"][1]
            raise Exception(
                "Sink not reachable for resource {}".format(unreachable_res))
        G.remove_nodes_from(nodes_to_remove)
        log.info("Removed {}/{} nodes".format(len(nodes_to_remove), n_nodes))
    return G


def preprocess_graph(
    G,
    max_res,
    min_res,
    preprocess,
    REF=None,
):
    """
    Applies preprocessing that removes nodes that cannot be reached due to
    resource limits.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats, optional
        :math:`[L, M_1, M_2, ..., M_{n\_res}]`
        upper bound for resource usage.
        We must have ``len(max_res)`` :math:`\geq 2`

    min_res : list of floats, optional
        :math:`[U, L_1, L_2, ..., L_{nres}]` lower bounds for resource usage.
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    preprocess : bool
        enables preprocessing routine.

    :return: If ``preprocess`` is True, returns the preprocessed graph if no
        exceptions are raised.

    """
    if REF:
        # Cannot apply pruning with custom REFs
        return deepcopy(G.copy())
    if preprocess:
        G = prune_graph(G, max_res, min_res)
        check(G)
    return deepcopy(G.copy())



import types


class Label(object):
    """
    Label object that allows comparison and the modelling of dominance
    relations.

    Parameters
    -----------

    weight : float
        cumulative edge weight

    node : string
        name of last node visited

    res : list
        cumulative edge resource consumption

    path : list
        all nodes in the path
    """

    _REF = None

    def __init__(self, weight, node, res, path):
        self.weight = weight
        self.node = node
        self.res = res
        self.path = path

    def __repr__(self):
        return str(self)

    def __str__(self):  # for printing purposes
        return "Label({0},{1},{2})".format(self.weight, self.node, self.res)

    def dominates(self, other, direction):
        # Determine whether self dominates other. Returns bool
        if self.node != other.node:
            raise Exception("Non-comparable labels given")
        else:
            # Assume self dominates other
            if self.weight > other.weight:
                return False
            if direction == "forward":
                if any(self.res > other.res):
                    return False
            elif direction == "backward":
                # Check for the monotone resource (non-increasing)
                if self.res[0] < other.res[0]:
                    return False
                # Check for all other resources (non-decreasing)
                if any(self.res[1:] > other.res[1:]):
                    return False
            return True

    def get_new_label(self, edge, direction):
        path = list(self.path)
        weight, res = edge[2]["weight"], edge[2]["res_cost"]
        node = edge[1] if direction == "forward" else edge[0]
        if node in path:  # If node already visited.
            return None
        else:
            path.append(node)
        if isinstance(self._REF, types.BuiltinFunctionType):
            res_new = self.res + res
        else:
            res_new = self._REF(self.res, edge)
        if direction == "backward":
            # Update monotone resource
            res_new[0] = self.res[0] - 1
        _new_label = Label(weight + self.weight, node, res_new, path)
        if _new_label == self:
            # If resulting label is the same
            return None
        return _new_label

    def feasibility_check(self, max_res, min_res):
        return all(max_res >= self.res) and all(min_res <= self.res)



"""
Adapted from https://github.com/100/Solid/blob/master/Solid/ParticleSwarm.py
"""

from __future__ import absolute_import
from __future__ import print_function

from math import sqrt
from abc import ABCMeta
from logging import getLogger
from numpy.random import RandomState
from numpy import (argmin, array, copy, diag_indices_from, exp, dot, zeros,
                   ones, where)

# Local imports
from cspy.algorithms.path_base import PathBase

log = getLogger(__name__)


class PSOLGENT(PathBase):
    """
    Particle Swarm Optimization with combined Local and Global Expanding
    Neighborhood Topology (PSOLGENT) algorithm for the (resource)
    constrained shortest path problem (`Marinakis et al 2017`_).

    Given the nature of our problem we have set the default parameters of
    the algorithm as suggested in the paper mentioned.

    Code adapted from `Solid`_.

    .. _Marinakis et al 2017: https://www.sciencedirect.com/science/article/pii/S0377221717302357z

    .. _Solid: https://github.com/100/Solid/blob/master/Solid/ParticleSwarm.py

    Parameters
    ----------
    G : object
        :class:`nx.Digraph()` must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage.

    min_res : list of floats
        :math:`[L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage.

    REF : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default : additive.

    preprocess : bool, optional
        enables preprocessing routine. Default : False.

    max_iter : int, optional
        Maximum number of iterations for algorithm. Default : 100.

    swarm_size : int, optional
        number of members in swarm. Default : 50.

    member_size : int, optional
        number of components per member vector. Default : ``len(G.nodes())``.

    neighbourhood_size : int, optional
        size of neighbourhood. Default : 10.

    lower_bound : list of floats, optional
        list of lower bounds. Default : ``numpy.zeros(member_size)``
        (no nodes in path).

    upper_bound : list of floats, optional
        list of upper bounds. Default : ``numpy.ones(member_size)``
        (all nodes in path).

    c1 : float, optional
        constant for 1st term in the velocity equation.
        Default : 1.35.

    c2 : float, optional
        contsant for 2nd term in the velocity equation.
        Default : 1.35.

    c3 : float, optional
        constant for 3rd term in the velocity equation.
        Default : 1.4.

    seed : None or int or numpy.random.RandomState instance, optional
        seed for PSOLGENT class. Default : None (which gives a single value
        numpy.random.RandomState).

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs

    Returns
    -------
    path : list
        nodes in resource feasible shortest path obtained.

    Raises
    ------
    Exception
        if no resource feasible path is found

    Notes
    -----
    The input graph must have a ``n_res`` attribute.
    The edges in the graph must all have a ``res_cost`` attribute.
    Also, we must have ``len(min_res)`` :math:`=` ``len(max_res)``.
    See `Using cspy`_.

    This algorithm requires a consistent sorting of the nodes in the graph.
    Please see comments and edit the function ``_sort_nodes`` accordingly.

    .. _Using cspy: https://cspy.readthedocs.io/en/latest/how_to.html

    Example
    -------
    .. code-block:: python

        >>> from cspy import PSOLGENT
        >>> from networkx import DiGraph
        >>> from numpy import zeros, ones, array
        >>> G = DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('Source', 'B', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('Source', 'C', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('A', 'C', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('A', 'E', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('A', 'F', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('B', 'C', res_cost=array([2, 1]), weight=-1)
        >>> G.add_edge('B', 'F', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('B', 'E', res_cost=array([10, 1]), weight=10)
        >>> G.add_edge('C', 'D', res_cost=array([1, 1]), weight=-1)
        >>> G.add_edge('D', 'E', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('D', 'F', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('D', 'Sink', res_cost=array([10, 10]), weight=10)
        >>> G.add_edge('F', 'Sink', res_cost=array([10, 1]), weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=array([1, 1]), weight=1)
        >>> n_nodes = len(G.nodes())
        >>> psolgent = PSOLGENT(G, [5, 5], [0, 0],
                                max_iter=200,
                                swarm_size=50,
                                member_size=n_nodes,
                                neighbourhood_size=50)
        >>> psolgent.run()
        >>> print(psolgent.path)
        ['Source', 'A', 'C', 'D', 'E', 'Sink']

    """

    __metaclass__ = ABCMeta

    def __init__(self,
                 G,
                 max_res,
                 min_res,
                 REF=None,
                 preprocess=False,
                 max_iter=100,
                 swarm_size=50,
                 member_size=None,
                 lower_bound=None,
                 upper_bound=None,
                 neighbourhood_size=10,
                 c1=1.35,
                 c2=1.35,
                 c3=1.4,
                 seed=None):
        # Pass arguments to parent class
        super().__init__(G, max_res, min_res, REF, preprocess)
        # Inputs
        self.swarm_size = swarm_size
        self.member_size = member_size if member_size else len(G.nodes())
        self.hood_size = neighbourhood_size
        self.lower_bound = zeros(member_size)
        self.upper_bound = ones(member_size)
        self.c1 = float(c1)
        self.c2 = float(c2)
        self.c3 = float(c3)
        self.max_iter = max_iter
        # PSO Specific Parameters
        self.iter = 0
        self.pos = None
        self.vel = None
        self.best = None
        self.fitness = None
        self.best_fit = None
        self.local_best = None
        self.global_best = None
        # Check seed and init `random_state`. Altered from:
        # https://github.com/scikit-learn/scikit-learn/blob/92af3dabbb5f3381a656f7727171f332b8928e05/sklearn/utils/validation.py#L764-L782
        if seed is None:
            self.random_state = RandomState()
        elif isinstance(seed, int):
            self.random_state = RandomState(seed)
        elif isinstance(seed, RandomState):
            self.random_state = seed
        else:
            raise Exception("{} cannot be used to seed".format(seed))

    def run(self):
        """
        Calculate shortest path with resource constraints.
        """
        self._init_swarm()
        while self.iter < self.max_iter:
            pos_new = self.pos + self._get_vel()
            self._update_best(self.pos, pos_new)
            self.pos = pos_new
            self.fitness = self._get_fitness(self.pos)
            self._global_best()
            self._local_best(self.iter, self.hood_size)
            if self.iter % 100 == 0:
                log.info("Iteration: {0}. Current best fit: {1}".format(
                    self.iter, self.best_fit))
            self.iter += 1
        if self.best_path:
            pass
        else:
            raise Exception("No resource feasible path has been found")

    def _init_swarm(self):
        # Initialises the variables that are altered during the algorithm
        self.pos = self.random_state.uniform(self.lower_bound,
                                             self.upper_bound,
                                             size=(self.swarm_size,
                                                   self.member_size))
        self.vel = self.random_state.uniform(
            self.lower_bound - self.upper_bound,
            self.upper_bound - self.lower_bound,
            size=(self.swarm_size, self.member_size))
        self.fitness = self._get_fitness(self.pos)
        self.best = copy(self.pos)
        self._global_best()
        self.local_best = copy(self.pos)

    def _get_vel(self):
        # Generate random numbers
        u1 = zeros((self.swarm_size, self.swarm_size))
        u1[diag_indices_from(u1)] = list(
            self.random_state.uniform(0, 1) for _ in range(self.swarm_size))
        u2 = zeros((self.swarm_size, self.swarm_size))
        u2[diag_indices_from(u2)] = list(
            self.random_state.uniform(0, 1) for _ in range(self.swarm_size))
        u3 = zeros((self.swarm_size, self.swarm_size))
        u3[diag_indices_from(u2)] = list(
            self.random_state.uniform(0, 1) for _ in range(self.swarm_size))
        # Coefficients
        c = self.c1 + self.c2 + self.c3
        chi_1 = 2 / abs(2 - c - sqrt(pow(c, 2) - 4 * c))
        # Returns velocity
        return (chi_1 * (self.vel + (self.c1 * dot(u1,
                                                   (self.best - self.pos)))) +
                (self.c2 * dot(u2, (self.global_best - self.pos))) +
                (self.c3 * dot(u3, (self.local_best - self.pos))))

    def _update_best(self, old, new):
        # Updates the best objective function values for each member
        old_fitness = array(self._get_fitness(old))
        new_fitness = array(self._get_fitness(new))
        best = zeros(shape=old.shape)
        if any(of < nf for of, nf in zip(old_fitness, new_fitness)):
            # replace indices in best with old members if lower fitness
            idx_old = list(
                idx for idx, val in enumerate(zip(old_fitness, new_fitness))
                if val[0] < val[1])
            best[idx_old] = old[idx_old]
            idx_new = where(best == 0)
            # replace indices in best with new members if lower fitness
            best[idx_new] = new[idx_new]
        else:
            best = new
        self.best = array(best)

    def _global_best(self):
        # Updates the global best across swarm
        if not self.best_fit or self.best_fit > min(self.fitness):
            self.global_best = array([self.pos[argmin(self.fitness)]] *
                                     self.swarm_size)
            self.best_fit = min(self.fitness)  # update best fitness

    def _local_best(self, i, hood_size):
        # Updates the local best across swarm
        bottom = max(i - hood_size, 0)
        top = min(bottom + hood_size, len(self.fitness))
        if top - bottom < hood_size:
            # Maximum length reached
            bottom = top - hood_size
        self.local_best = array([self.pos[argmin(self.fitness[bottom:top])]] *
                                self.swarm_size)

    # Fitness #
    # Fitness conversion to path representation of solutions and evaluation
    def _get_fitness(self, pos):
        # Applies objective function to all members of swarm
        return list(map(self._evaluate_member, pos))

    def _evaluate_member(self, member):
        rand = self.random_state.uniform(0, 1)
        self._update_current_nodes(self._discretise_solution(member, rand))
        return self._get_fitness_member()

    @staticmethod
    def _discretise_solution(member, rand):
        sig = array(1 / (1 + exp(-member)))
        return array([1 if s < rand else 0 for s in sig])

    def _update_current_nodes(self, arr):
        """
        Saves binary representation of nodes in path.
        0 not present, 1 present.
        """
        nodes = self._sort_nodes(list(self.G.nodes()))
        self.current_nodes = list(
            nodes[i] for i in range(len(nodes)) if arr[i] == 1)

    def _get_fitness_member(self):
        # Returns the objective for a given path
        return self._fitness(self.current_nodes)

    def _fitness(self, nodes):
        edges = self._get_edges(nodes)
        disconnected, path = self._check_edges(edges)
        if edges:
            if disconnected:
                return 1e7  # disconnected path
            cost = self._check_path(path, edges)
            if cost is not False:
                # Valid path with penalty
                return cost
            else:
                return 1e6  # path not valid
        else:
            return 1e6  # no path

    # Path-related methods #
    def _get_edges(self, nodes):
        # Creates a list of edges given the nodes selected
        return list(
            edge for edge in self.G.edges(self.G.nbunch_iter(nodes), data=True)
            if edge[0:2] in zip(nodes, nodes[1:]))

    @staticmethod
    def _check_edges(edges):
        """
        If edges given, saves the path provided.
        Returns whether the path is disconnected or not
        """
        if edges:
            path = [edge[0] for edge in edges]
            path.append(edges[-1][1])
            return any(edge[1] not in path for edge in edges), path
        else:
            return None, None

    def _check_path(self, path, edges):
        """
        Returns False if path is not valid
        Penalty/cost otherwise
        """
        if path:
            if len(path) > 2 and (path[0] == 'Source' and path[-1] == 'Sink'):
                base_cost = sum(edge[2]['weight'] for edge in edges)
                self.st_path = path
                if self.check_feasibility() is True:
                    log.debug("Resource feasible path found")
                    return base_cost
                else:
                    # penalty for resource infeasible valid path
                    return 1e5 + base_cost
            else:
                return False
        else:
            return False

    @staticmethod
    def _sort_nodes(nodes):
        """
        Sort nodes between Source and Sink. If node data allows sorting,
        edit the line below to pass `list(self.G.nodes(data=True))` and used that
        in the sorting function below.

        For example, if nodes have an attribute `pos` a tuple that contains
        the position of a node in space (x, y), replace the return with:
        `return sorted(nodes, key=lambda n: n[1]['pos'][0])`
        edit the sorting function to make use of it.
        """
        order_dict = {'Source': '0', 'Sink': 'Z{}'.format(len(nodes))}
        return sorted(nodes,
                      key=lambda x: order_dict[x] if x in order_dict else x)



from __future__ import absolute_import
from __future__ import print_function

from logging import getLogger
from collections import deque
from networkx import NetworkXException

# Local imports
from cspy.algorithms.path_base import PathBase

log = getLogger(__name__)


class Tabu(PathBase):
    """
    Simple Tabu-esque algorithm for the (resource) constrained shortest
    path problem.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).

    min_res : list of floats
        :math:`[L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)``.

    REF : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default : additive.

    preprocess : bool, optional
        enables preprocessing routine. Default : False.

    max_depth : int, optional
        depth for search of shortest simple path. Default : 1000.
        If the total number of simple paths is less than max_depth,
        then the shortest path is used.

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs


    Notes
    -----
    The input graph must have a ``n_res`` attribute.
    The edges in the graph must all have a ``res_cost`` attribute.
    See `Using cspy`_

    .. _Using cspy: https://cspy.readthedocs.io/en/latest/how_to.html

    Example
    -------
    To run the algorithm, create a :class:`Tabu` instance, call `run`, and then
    query the attributes of interest: `path`, `total_cost`, or
    `consumed_resources`.

    .. code-block:: python

        >>> from cspy import Tabu
        >>> from networkx import DiGraph
        >>> from numpy import array
        >>> G = DiGraph(directed=True, n_res=2)
        >>> G.add_edge('Source', 'A', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('Source', 'B', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('A', 'C', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('B', 'C', res_cost=array([2, 1]), weight=-1)
        >>> G.add_edge('C', 'D', res_cost=array([1, 1]), weight=-1)
        >>> G.add_edge('D', 'E', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('D', 'F', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('F', 'Sink', res_cost=array([1, 1]), weight=1)
        >>> G.add_edge('E', 'Sink', res_cost=array([1, 1]), weight=1)
        >>> max_res, min_res = [5, 5], [0, 0]
        >>> tabu = Tabu(G, max_res, min_res)
        >>> tabu.run()
        >>> print(tabu.path)
        ['Source', 'A', 'C', 'D', 'E', 'Sink']

    """

    def __init__(self,
                 G,
                 max_res,
                 min_res,
                 REF=None,
                 preprocess=False,
                 max_depth=1000):
        # Pass arguments to SimplePath object
        super().__init__(G, max_res, min_res, REF, preprocess)
        # Algorithm specific parameters
        self.max_depth = max_depth
        self.iteration = 0
        self.stop = False
        self.neighbour = 'Source'
        self.neighbourhood = list()
        self.tabu_edge = None
        self.edges_to_check = dict(self.G.edges())

    def run(self):
        """
        Calculate shortest path with resource constraints.
        """
        while self.stop is False:
            self._algorithm()
            self.iteration += 1

        if self.best_path:
            pass
        else:
            raise Exception("No resource feasible path has been found")

    def _algorithm(self):
        path = []
        try:
            path = self.update_simple_path(self.neighbour, self.max_depth)
        except NetworkXException:
            pass
        if path:
            self._update_path(self.neighbour, path)
            edge_or_true = self.check_feasibility()
            # If there is a resource feasible path
            if edge_or_true is True:
                self.stop = True
            # Otherwise, use the infeasible edge as the next tabu edge
            else:
                self._get_neighbour(edge_or_true)
        else:
            log.debug("No path found")
            self._get_neighbour(self.tabu_edge)

    # Path-related methods #
    def _update_path(self, neighbour, path):
        # Joins path using previous path and [neighbour, ..., sink] path
        if neighbour == "Source":
            self.st_path = path
        elif neighbour in self.st_path:
            # Paths can be joined at neighbour
            self.st_path = list(node for node in self.st_path if
                                (node != neighbour and self.st_path.index(node)
                                 < self.st_path.index(neighbour))) + path
        else:
            self._merge_paths(neighbour, path)

    def _merge_paths(self, neighbour, path):
        branch_path = [n for n in self.st_path if n not in path]
        for node in reversed(branch_path):
            if (node, neighbour) in self.G.edges():
                self.st_path = list(n for n in branch_path if (
                    branch_path.index(n) <= branch_path.index(node))) + path
                break

    # Algorithm-specific methods #
    def _update_tabu_edge(self, edge):
        # If a tabu edge has already been selected
        if self.tabu_edge:
            # Revert old tabu edge weight to original
            self.add_edge_back(self.tabu_edge)
        # Replace new tabu edge weight with large number
        self.remove_edge(edge)
        self.tabu_edge = edge

    def _get_neighbour(self, edge=None):
        """
        Get next neighbour to the resource infeasible edge.
        Update the tabu edge.
        """
        if self.edges_to_check:
            if edge and edge[:2] in self.edges_to_check:
                # If edge not already been seen
                current_edge = edge
            else:
                current_edge = self._get_next_neighbour_edge(self.tabu_edge)
        else:
            self.stop = True
            return
        self.edges_to_check.pop(current_edge[:2], None)
        self.neighbour = current_edge[0]
        self._update_tabu_edge(current_edge)

    def _get_next_neighbour_edge(self, edge):
        # Retrieves the edge adjacent to node with the greatest weight
        # If neighbourhood doesn't exist
        if not self.neighbourhood:
            node = edge[0]
            if node == "Source":
                self.stop = True
                return edge
            else:
                nodes_iter = self.G.predecessors(node)
                self.neighbourhood = deque(e for e in self.G.edges(
                    self.G.nbunch_iter(nodes_iter), data=True)
                                           if e[1] == node and e != edge)
        # Get the edge in the neighbourhood with greatest weight
        next_edge = max(self.neighbourhood, key=lambda x: x[2]['weight'])
        # delete edge from neighbourhood
        del self.neighbourhood[self.neighbourhood.index(next_edge)]
        return next_edge



from __future__ import absolute_import

from copy import deepcopy
from itertools import repeat
from operator import add
from logging import getLogger
from collections import OrderedDict, deque

from numpy import array
from numpy.random import RandomState

# Local module imports
from cspy.checking import check
from cspy.algorithms.label import Label
from cspy.preprocessing import preprocess_graph

log = getLogger(__name__)


class BiDirectional:
    """
    Implementation of the bidirectional labeling algorithm with dynamic
    half-way point (`Tilk 2017`_).
    This requires the half-way procedure from `Righini and Salani (2006)`_,
    also implemented.
    Depending on the range of values for U, L, we get
    four different algorithms. See self.name_algorithm and Notes.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats
        :math:`[H_F, M_1, M_2, ..., M_{n\_res}]` upper bounds for resource
        usage (including initial forward stopping point).
        We must have ``len(max_res)`` :math:`\geq 2`.

    min_res : list of floats
        :math:`[H_B, L_1, L_2, ..., L_{n\_res}]` lower bounds for resource
        usage (including initial backward stopping point).
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    REF : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default : additive.

    preprocess : bool, optional
        enables preprocessing routine. Default : False.

    direction : string, optional
        preferred search direction.
        Either "both", "forward", or, "backward". Default : "both".

    method : string, optional
        preferred method for determining search direction.
        Either "random", "generated" (direction with least number of generated labels),
        "processed" (direction with least number of processed labels), or,
        "unprocessed" (direction with least number of unprocessed labels).
        Default: "random".

    seed : None or int or numpy.random.RandomState instance, optional
        seed for PSOLGENT class. Default : None (which gives a single value
        numpy.random.RandomState).

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs

    Notes
    -----
    The input graph must have a ``n_res`` attribute which must be
    :math:`\geq 2`. The edges in the graph must all have a ``res_cost``
    attribute.

    According to the inputs, four different algorithms can be used.
    If you'd like to check which algorithm is running given the resource
    limits, call the method :func:`BiDirectional.name_algorithm()`
    for a log with the classification.

    - ``direction`` = "forward": Monodirectional forward labeling algorithm
    - :math:`H_F == H_B`: Bidirectional labeling algorithm with static halfway point.
    - ``direction`` = "backward": Monodirectional backward labeling algorithm
    - :math:`H_F > H_B`: Bidirectional labeling algorithm with dynamic halfway point.
    - :math:`H_F < H_B`: The algorithm won't go anywhere!

    Example
    -------
    To run the algorithm, create a :class:`BiDirectional` instance and call
    ``run``.

    .. code-block:: python

        >>> from cspy import BiDirectional
        >>> from networkx import DiGraph
        >>> from numpy import array
        >>> G = DiGraph(directed=True, n_res=2)
        >>> G.add_edge("Source", "A", res_cost=array([1, 2]), weight=0)
        >>> G.add_edge("A", "B", res_cost=array([1, 0.3]), weight=0)
        >>> G.add_edge("A", "C", res_cost=array([1, 0.1]), weight=0)
        >>> G.add_edge("B", "C", res_cost=array([1, 3]), weight=-10)
        >>> G.add_edge("B", "Sink", res_cost=array([1, 2]), weight=10)
        >>> G.add_edge("C", "Sink", res_cost=array([1, 10]), weight=0)
        >>> max_res, min_res = [4, 20], [1, 0]
        >>> bidirec = BiDirectional(G, max_res, min_res, direction="both")
        >>> bidirec.run()
        >>> print(bidirec.path)
        ["Source", "A", "B", "C", "Sink"]

    .. _Tilk 2017: https://www.sciencedirect.com/science/article/pii/S0377221717302035
    .. _Righini and Salani (2006): https://www.sciencedirect.com/science/article/pii/S1572528606000417
    """

    def __init__(self,
                 G,
                 max_res,
                 min_res,
                 REF=None,
                 preprocess=False,
                 direction="both",
                 method="random",
                 seed=None):
        # Check inputs and preprocess G unless option disabled
        check(G, max_res, min_res, REF, direction, __name__)
        # Preprocess graph
        self.G = preprocess_graph(G, max_res, min_res, preprocess, REF)
        self.direc_in = direction
        self.max_res, self.min_res = max_res.copy(), min_res.copy()
        self.max_res_in, self.min_res_in = array(max_res.copy()), array(
            min_res.copy())
        self.method = method
        # To expose results
        self.best_label = None

        # Algorithm specific parameters #
        # set bounds for bacward search
        bwd_start = deepcopy(min_res)
        bwd_start[0] = max_res[0]
        # Current forward and backward labels
        self.current_label = OrderedDict({
            "forward": Label(0, "Source", min_res, ["Source"]),
            "backward": Label(0, "Sink", bwd_start, ["Sink"])
        })
        # Unprocessed labels dict (both directions)
        self.unprocessed_labels = OrderedDict({
            "forward": deque(),
            "backward": deque()
        })
        # All generated label
        self.generated_labels = OrderedDict({"forward": 0, "backward": 0})
        # To save all best labels
        self.best_labels = OrderedDict({
            "forward": deque(),
            "backward": deque()
        })
        # Final labels dicts for unidirectional search
        self.final_label = None

        # If given, set REFs for dominance relations and feasibility checks
        if REF:
            Label._REF = REF
        else:
            Label._REF = add
        # Init with seed if given
        if seed is None:
            self.random_state = RandomState()
        elif isinstance(seed, int):
            self.random_state = RandomState(seed)
        elif isinstance(seed, RandomState):
            self.random_state = seed
        else:
            raise Exception("{} cannot be used to seed".format(seed))

    def run(self):
        """
        Calculate shortest path with resource constraints.
        """
        while self.current_label["forward"] or self.current_label["backward"]:
            direc = self._get_direction()
            if direc:
                self._algorithm(direc)
            else:
                break
        return self._process_paths()

    @property
    def path(self):
        """
        Get list with nodes in calculated path.
        """
        if not self.best_label:
            raise Exception("Please call the .run() method first")
        return self.best_label.path

    @property
    def total_cost(self):
        """
        Get accumulated cost along the path.
        """
        if not self.best_label:
            raise Exception("Please call the .run() method first")
        return self.best_label.weight

    @property
    def consumed_resources(self):
        """
        Get accumulated resources consumed along the path.
        """
        if not self.best_label:
            raise Exception("Please call the .run() method first")
        return self.best_label.res

    ###########################
    # Classify Algorithm Type #
    ###########################
    def name_algorithm(self):
        """
        Determine which algorithm is running.
        """
        HF, HB = self.max_res[0], self.min_res[0]
        if self.direc_in == "forward":
            log.info("Monodirectional forward labeling algorithm")
        elif self.direc_in == "backward":
            log.info("Monodirectional backward labeling algorithm")
        elif HF > HB:
            log.info("Bidirectional labeling algorithm with" +
                     " dynamic halfway point")
        else:
            log.info("The algorithm can't move in either direction!")

    #############
    # DIRECTION #
    #############
    def _get_direction(self):
        if self.direc_in == "both":
            if (self.current_label["forward"] and
                    not self.current_label["backward"]):
                return "forward"
            elif (not self.current_label["forward"] and
                  self.current_label["backward"]):
                return "backward"
            elif (self.current_label["forward"] and
                  self.current_label["backward"]):
                if self.method == "random":
                    # return a random direction
                    return self.random_state.choice(["forward", "backward"])
                elif self.method == "generated":
                    # return direction with least number of generated labels
                    return ("forward" if self.generated_labels["forward"] <
                            self.generated_labels["backward"] else "backward")
                elif self.method == "processed":
                    # return direction with least number of "processed" labels
                    return ("forward" if len(self.best_labels["forward"]) < len(
                        self.best_labels["backward"]) else "backward")
                elif self.method == "unprocessed":
                    # return direction with least number of unprocessed_labels labels
                    return ("forward" if len(self.unprocessed_labels["forward"])
                            < len(self.unprocessed_labels["backward"]) else
                            "backward")
            else:  # if both are empty
                return
        else:
            if (not self.current_label["forward"] and
                    not self.current_label["backward"]):
                return
            elif not self.current_label[self.direc_in]:
                return
            else:
                return self.direc_in

    #############
    # ALGORITHM #
    #############
    def _algorithm(self, direc):
        if direc == "forward":  # forward
            idx = 0  # index for head node
            # Update backwards half-way point
            self.min_res[0] = max(
                self.min_res[0],
                min(self.current_label[direc].res[0], self.max_res[0]))
        else:  # backward
            idx = 1  # index for tail node
            # Update forwards half-way point
            self.max_res[0] = min(
                self.max_res[0],
                max(self.current_label[direc].res[0], self.min_res[0]))
        # Select edges with the same head/tail node as the current label node.
        edges = deque(e for e in self.G.edges(data=True)
                      if e[idx] == self.current_label[direc].node)
        # If Label not been seen before, initialise a list
        # Propagate current label along all suitable edges in current direction
        deque(map(self._propagate_label, edges, repeat(direc, len(edges))))
        # Extend label
        next_label = self._get_next_label(direc)
        # Update current label
        self.current_label[direc] = next_label
        # Dominance checks
        self._check_dominance(next_label, direc)

    def _propagate_label(self, edge, direc):
        # Label propagation #
        new_label = self.current_label[direc].get_new_label(edge, direc)
        # If the new label is resource feasible
        if new_label and new_label.feasibility_check(self.max_res,
                                                     self.min_res):
            # And is not already in the unprocessed labels list
            if (new_label not in self.unprocessed_labels[direc]):
                self.unprocessed_labels[direc].append(new_label)

    def _get_next_label(self, direc):
        # Label Extension #
        # Add current label to processed list.
        current_label = self.current_label[direc]
        unproc_labels = self.unprocessed_labels[direc]

        # Add 1 to count of generated_labels
        self.generated_labels[direc] += 1
        self._remove_labels([current_label], direc, unproc=True)
        # Return label with minimum monotone resource for the forward search
        # and the maximum monotone resource for the backward search
        if unproc_labels:
            if direc == "forward":
                return min(unproc_labels, key=lambda x: x.res[0])
            else:
                return max(unproc_labels, key=lambda x: x.res[0])
        else:
            return None

    #############
    # DOMINANCE #
    #############
    def _check_dominance(self, label_to_check, direc, unproc=True, best=False):
        """
        For all labels, checks if ``label_to_check`` is dominated,
        or itself dominates any other label in either the unprocessed_labels
        list or the non-dominated labels list.
        If this is found to be the case, the dominated label(s) is(are)
        removed from the appropriate list.
        """
        # Select appropriate list to check
        if unproc:
            labels_to_check = self.unprocessed_labels[direc]
        elif best:
            labels_to_check = self.best_labels[direc]
        # If label is not None (at termination)
        if label_to_check:
            labels_to_pop = deque()
            # Gather all comparable labels (same node)
            all_labels = deque(
                l for l in labels_to_check
                if l.node == label_to_check.node and l != label_to_check)
            # Add to list for removal if they are dominated
            labels_to_pop.extend(
                l for l in all_labels if label_to_check.dominates(l, direc))
            # Add input label for removal if itself is dominated
            if any(l.dominates(label_to_check, direc) for l in all_labels):
                labels_to_pop.append(label_to_check)
            elif unproc:
                # check and save current label
                self._save_current_best_label(direc)
            # if unprocessed labels checked then remove labels_to_pop
            if unproc:
                self._remove_labels(labels_to_pop, direc, unproc, best)
            # Otherwise, return labels_to_pop for later removal
            elif best:
                return labels_to_pop

    def _remove_labels(self, labels_to_pop, direc, unproc=True, best=False):
        """
        Remove all labels in ``labels_to_pop`` from either the array of
        unprocessed labels or the array of non-dominated labels
        """
        # Remove all processed labels from unprocessed dict
        for label_to_pop in deque(set(labels_to_pop)):
            if unproc and label_to_pop in self.unprocessed_labels[direc]:
                idx = self.unprocessed_labels[direc].index(label_to_pop)
                del self.unprocessed_labels[direc][idx]
            elif best and label_to_pop in self.best_labels[direc]:
                idx = self.best_labels[direc].index(label_to_pop)
                del self.best_labels[direc][idx]

    def _save_current_best_label(self, direc):
        """
        Label saving
        """
        current_label = self.current_label[direc]
        final_label = self.final_label

        if self.direc_in == "both":
            self.best_labels[direc].append(current_label)
            flip_direc = "forward" if direc == "backward" else "backward"
            # Check if other direction traversed
            if self.best_labels[flip_direc]:
                return
        # If first label
        if not final_label:
            self.final_label = current_label
            return
        # Otherwise, check dominance and replace
        try:
            if self._full_dominance_check(current_label, final_label, direc):
                log.debug("Saving {} as best, with path {}".format(
                    current_label, current_label.path))
                self.final_label = current_label
        except Exception:
            # Labels are not comparable i.e. Belong to different nodes
            if (direc == "forward" and (current_label.path[-1] == "Sink" or
                                        final_label.node == "Source")):
                log.debug("Saving {} as best, with path {}".format(
                    current_label, current_label.path))
                self.final_label = current_label
            elif (direc == "backward" and (current_label.path[-1] == "Source" or
                                           final_label.node == "Sink")):
                log.debug("Saving {} as best, with path {}".format(
                    current_label, current_label.path))
                self.final_label = current_label

    @staticmethod
    def _full_dominance_check(label1, label2, direc):
        """
        Checks whether label 1 dominates label 2 for the input direction.
        In the case when neither dominates , i.e. they are non-dominated,
        the direction is flipped labels are compared again.
        """
        label1_dominates = label1.dominates(label2, direc)
        label2_dominates = label2.dominates(label1, direc)
        # label1 dominates label2 for the input direction
        if label1_dominates:
            return True
        # Both non-dominated labels in this direction.
        elif (not label1_dominates and not label2_dominates):
            # flip directions
            flip_direc = "forward" if direc == "backward" else "backward"
            label1_dominates_flipped = label1.dominates(label2, flip_direc)
            # label 1 dominates label2 in the flipped direction
            if label1_dominates_flipped:
                return True
            elif label1.weight < label2.weight:
                return True

    ###################
    # PATH PROCESSING #
    ###################
    def _process_paths(self):
        # Processing of output path.
        if (self.direc_in == "both" and self.best_labels["forward"] and
                self.best_labels["backward"]):
            # If bi-directional algorithm used, run path joining procedure.
            self._clean_up_best_labels()
            self._join_paths()
        else:
            # If mono-directional algorithm used or both directions not traversed,
            # return the appropriate path
            if (self.direc_in == "forward" or
                (self.direc_in == "both" and not self.best_labels["backward"])):
                # Forward
                self.best_label = self.final_label
            else:
                # Backward
                self.best_label = self._process_bwd_label(
                    self.final_label, self.min_res_in)

    def _process_bwd_label(self, label, cumulative_res):
        # Reverse backward path and inverts resource consumption
        label.path.reverse()
        label.res[0] = self.max_res_in[0] - label.res[0]
        label.res = label.res + cumulative_res
        return label

    def _clean_up_best_labels(self):
        # Removed all dominated labels in best_labels
        for direc in ["forward", "backward"]:
            labels_to_pop = deque()
            for l in self.best_labels[direc]:
                labels_to_pop.extend(
                    self._check_dominance(l, direc, unproc=False, best=True))
            self._remove_labels(labels_to_pop, direc, unproc=False, best=True)

    def _join_paths(self):
        """
        The procedure "Join" or Algorithm 3 from `Righini and Salani (2006)`_.

        Modified to get rid of nested for loops and reduced search.

        :return: list with the final path.

        .. _Righini and Salani (2006): https://www.sciencedirect.com/science/article/pii/S1572528606000417
        """
        log.debug("joining")
        for fwd_label in self.best_labels["forward"]:
            # Create generator for backward labels for current forward label.
            # Includes only those that:
            # 1. Paths can be joined (exists a connecting edge)
            # 2. Introduces no cycles
            # 3. When combined with the forward label, they satisfy the halfway check
            bwd_labels = (l for l in self.best_labels["backward"] if (
                (fwd_label.node, l.node) in self.G.edges() and \
                not any(n in fwd_label.path for n in l.path) and
                self._half_way(fwd_label, l)))
            for bwd_label in bwd_labels:
                # Merge two labels
                merged_label = self._merge_labels(fwd_label, bwd_label)
                # Check resource feasibility
                if (merged_label and merged_label.feasibility_check(
                        self.max_res_in, self.min_res_in)):
                    # Save label
                    self._save(merged_label)

    def _half_way(self, fwd_label, bwd_label):
        """
        Half-way check from `Righini and Salani (2006)`_.
        Checks if a pair of labels is closes to the half-way point.

        :return: bool. True if the half-way check passes, false otherwise.

        .. _Righini and Salani (2006): https://www.sciencedirect.com/science/article/pii/S1572528606000417
        """
        phi = abs(fwd_label.res[0] - (self.max_res_in[0] - bwd_label.res[0]))
        if phi >= 0:
            return True
        else:
            return False

    def _merge_labels(self, fwd_label, bwd_label):
        """
        Merge labels produced by a backward and forward label.

        Paramaters
        ----------
        fwd_label : label.Label object
        bwd_label : label.Label object

        Returns
        -------
        merged_label : label.Label object
            If an s-t compatible path can be obtained the appropriately
            extended and merged label is returned

        None
            Otherwise.
        """
        # Make a copy of the backward label
        _bwd_label = deepcopy(bwd_label)
        # Reconstruct edge with edge data
        edge = (fwd_label.node, _bwd_label.node,
                self.G[fwd_label.node][_bwd_label.node])
        # Extend forward label along joining edge
        label = fwd_label.get_new_label(edge, "forward")

        if not label:
            return
        # Process backward label
        self._process_bwd_label(_bwd_label, label.res)
        # Record total weight, total_res and final path
        weight = fwd_label.weight + edge[2]['weight'] + _bwd_label.weight
        # Get total consumed resources (inverted)
        total_res = _bwd_label.res
        final_path = fwd_label.path + _bwd_label.path
        merged_label = Label(weight, "Sink", total_res, final_path)
        return merged_label

    def _save(self, label):
        # Saves a label for exposure
        if not self.best_label or self._full_dominance_check(
                label, self.best_label, "forward"):
            log.debug("Saving label {} as best".format(label))
            log.debug("With path {}".format(label.path))
            self.best_label = label



from networkx import DiGraph, NetworkXException, has_path
from numpy import ndarray


def check(G,
          max_res=None,
          min_res=None,
          REF=None,
          direction=None,
          algorithm=None):
    """
    Checks whether inputs and the graph are of the appropriate types and
    have the required properties.
    For non-specified REFs, removes nodes that cannot be reached due to
    resource limits.

    Parameters
    ----------
    G : object instance :class:`nx.Digraph()`
        must have ``n_res`` graph attribute and all edges must have
        ``res_cost`` attribute.

    max_res : list of floats, optional
        :math:`[L, M_1, M_2, ..., M_{n\_res}]`
        upper bound for resource usage.
        We must have ``len(max_res)`` :math:`\geq 2`

    min_res : list of floats, optional
        :math:`[U, L_1, L_2, ..., L_{nres}]` lower bounds for resource usage.
        We must have ``len(min_res)`` :math:`=` ``len(max_res)`` :math:`\geq 2`

    REF : function, optional
        Custom resource extension function. See `REFs`_ for more details.
        Default: additive.

    direction : string, optional
        preferred search direction. Either 'both','forward', or, 'backward'.
        Default : 'both'.

    .. _REFs : https://cspy.readthedocs.io/en/latest/how_to.html#refs

    :raises: Raises exceptions if incorrect input is given.
        If multiple exceptions are raised, an exception with a list of
        errors is raised.
    """
    errors = []
    if REF:
        # Cannot apply pruning with custom REFs
        try:
            _check_REFs(REF)
        except Exception as e:
            errors.append(e)
    # Select checks to perform based on the input provided
    if max_res and min_res and direction:
        check_funcs = [
            _check_res, _check_direction, _check_graph_attr, _check_edge_attr,
            _check_path
        ]
    elif max_res and min_res:
        check_funcs = [
            _check_res, _check_graph_attr, _check_edge_attr, _check_path
        ]
    else:
        check_funcs = [_check_path]
    # Check all functions in check_funcs
    for func in check_funcs:
        try:
            func(G, max_res, min_res, direction, algorithm)
        except Exception as e:
            errors.append(e)  # if check fails save error message
    if errors:
        # if any check has failed raise an exception with all the errors
        raise Exception('\n'.join('{}'.format(item) for item in errors))


def _check_res(G, max_res, min_res, direction, algorithm):
    if isinstance(max_res, list) and isinstance(min_res, list):
        if len(max_res) == len(min_res):
            if (algorithm and 'bidirectional' in algorithm and
                    len(max_res) < 2):
                raise TypeError("Resources must be of length >= 2")
            if (all(isinstance(i, (float, int)) for i in max_res) and
                    all(isinstance(i, (float, int)) for i in min_res)):
                pass
            else:
                raise TypeError("Elements of input lists must be numbers")
        else:
            raise TypeError("Input lists have to be equal length")
    else:
        raise TypeError("Inputs have to be lists with length >= 2")


def _check_direction(G, max_res, min_res, direction, algorithm):
    if direction not in ['forward', 'backward', 'both']:
        raise TypeError(
            "Input direction has to be 'forward', 'backward', or 'both'")


def _check_graph_attr(G, max_res, min_res, direction, algorithm):
    """Checks whether input graph has n_res attribute"""
    if isinstance(G, DiGraph):
        if 'n_res' not in G.graph:
            raise TypeError("Input graph must have 'n_res' attribute")
    else:
        raise TypeError("Input must be a nx.Digraph()")


def _check_edge_attr(G, max_res, min_res, direction, algorithm):
    """Checks whether edges in input graph have res_cost attribute"""
    if not all('res_cost' in edge[2] for edge in G.edges(data=True)):
        raise TypeError("Input graph must have edges with 'res_cost' attribute")
    if not all(
            len(edge[2]['res_cost']) == G.graph['n_res']
            for edge in G.edges(data=True)):
        raise TypeError(
            "Edges must have 'res_cost' attribute with length equal to 'n_res'")
    if not all(
            len(edge[2]['res_cost']) == len(max_res) == len(min_res)
            for edge in G.edges(data=True)):
        raise TypeError(
            "Edges must have 'res_cost' attribute with length equal to" +
            " 'min_res' == 'max_res")
    if not all(
            isinstance(edge[2]['res_cost'], ndarray)
            for edge in G.edges(data=True)):
        raise TypeError("The edge 'res_cost' attribute must be a numpy.array")


def _check_path(G, max_res, min_res, direction, algorithm):
    """Checks whether a 'Source' -> 'Sink' path exists and if there are
    negative edge cycles in the graph.
    Also covers nodes missing and other standard networkx exceptions"""
    try:
        if not has_path(G, 'Source', 'Sink'):
            raise NetworkXException("Disconnected Graph")
    except NetworkXException as e:
        raise Exception("An error occurred: {}".format(e))


def _check_REFs(REF):
    if REF and not callable(REF):
        raise TypeError("REF functions must be callable")
